/*
*   umasip is a Videoconference program that uses the SIP protocol (RFC 3261)
*   Copyright (C) 2004 Pablo Montoro Escao
*
*   This program is free software; you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation; either version 2 of the License, or
*   (at your option) any later version.
*
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with this program; if not, write to the Free Software
*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*/
import org.mitre.jsip.*;
import org.mitre.jsip.event.*;

/**
 * <p>Programa: umasip</p>
 * <p>Copyright: Copyright GNU/GPL (c) 2004 Pablo Montoro Escao</p>
 * @author Pablo Montoro Escao
 * @version 1.00
 * @since SDK 1.4.2.
 */

public class Agente {

    public final int AUDIO_PORT=23322; //Puerto por defecto para Audio
    public final int VIDEO_PORT=23326; //Puerto por defecto para Video
    private final String AUDIO_TRANSPORT="RTP/AVP"; //Transport Protocol Used
    private final String VIDEO_TRANSPORT="RTP/AVP"; //RTP Audio/Video Profile
    private final MimeContentType TIPO_SDP=new MimeContentType("application/sdp");

    private SipClient client;
    private SipCall call,closingCall,rejectingCall;
    private SipCallMember member;
    private SipUser localUser;

    public String audioPT=null; //RFC1890 Payload Type (null=NO TX AUDIO)
    public String videoPT=null; //RFC1890 Payload Type (null=NO TX VIDEO)
    public String remoto; //Direccion URI Remota
    public String ipRemoto; //Direccion IP Remota
    public int paRemoto; //Puerto de Audio Remoto
    public int pvRemoto; //Puerto de Video Remoto
    public String ipLocal; //Direccion IP Local
    public int paLocal; //Puerto de Audio Local
    public int pvLocal; //Puerto de Video Local

    public boolean solicitaLlamada=false; //Indica si llega una llamada
    public boolean solicitaColgar=false; //Indica si se solicita colgar llamada
    public int respuestaRemota=0; //tipo de respuesta recibida a una llamada

    private String sipDebug=""; //Informacion debug del protocolo SIP

    /**
     * Constructor para el Agente de Usuario SIP. Cliente y Servidor (UAC y UAS)
     * Se adopta como usuario local anonymous@direccion_IP_local
     */
    public Agente() {
        this("anonymous");
    }

    /**
     * Constructor para el Agente de Usuario SIP. Cliente y Servidor (UAC y UAS)
     * Se adopta como usuario local valor_de_usuario@direccion_IP_local
     * @param usuario Nombre del usuario
     */
    public Agente(String usuario) {
        String ip=Sip.getLocalAddress();
        init(usuario,ip);
    }

    /**
     * Constructor para el Agente de Usuario SIP. Cliente y Servidor (UAC y UAS)
     * Se adopta como usuario local valor_de_usuario@direccion_IP_local
     * @param usuario Nombre del usuario
     * @param ip Direccion IP local
     */
    public Agente(String usuario,String ip) {
        init(usuario,ip);
    }

    private void init(String usuario,String ip) {

        //Creamos el agente de usuario SIP (cliente UAC)
        client=SipClient.getSipClient();
        client.setDefaultUserMode(false);
        client.setExplicitProxyMode(false);
        client.setCallForward(false);
        client.setBusyMessage("busy!");

        ipLocal=ip;
        Sip.setLocalAddress(ipLocal);

        localUser=new SipUser(client,new SipUri(usuario+"@"+ipLocal));
        client.setContactUri(new SipUri(usuario+"@"+ipLocal));

        //Creamos el Listener (servidor UAS) para los mensajes entrantes
        CallListener cl=new CallListener() {
            public synchronized void responseReceived(CallEvent ce) {
                SipResponseMessage responseMessage=
                    (SipResponseMessage) ce.getMessage();
                int codigo=responseMessage.getStatus().getCode();
                String callID=responseMessage.getHeaderData(SipHeader.Call_ID);
                String string=""; //cadena temporal para obtener Payload Types

                if ((closingCall!=null) && (codigo==200) &&
                    (callID.equals(closingCall.getCallId()))) {

                    client.deleteCall(closingCall);
                    closingCall=null;
                    addTimedSIP(" <<==== Response OK (200) <<==== "+
                                responseMessage.getHeaderData(SipHeader.To)+"\n");

                } else if ((closingCall!=null) && (codigo==481) &&
                    (callID.equals(closingCall.getCallId()))) {

                    client.deleteCall(closingCall);
                    closingCall=null;
                    addTimedSIP(" <<==== Response NOT EXIST (481) <<==== "+
                                responseMessage.getHeaderData(SipHeader.To)+"\n");

                } else if ((call!=null) && (codigo==100) &&
                    (callID.equals(call.getCallId()))) {
                    addTimedSIP(" <<==== Response TRYING (100) <<==== "+remoto);
                } else if ((call!=null) && (codigo==180) &&
                    (callID.equals(call.getCallId()))) {
                    addTimedSIP(" <<==== Response RINGING (180) <<==== "+remoto);
                } else if ((call!=null) && (codigo==200) &&
                    (callID.equals(call.getCallId()))) {

                    addTimedSIP(" <<==== Response OK (200) <<==== "+remoto);
                    SdpMessage sdpMessage=
                        new SdpMessage(responseMessage.messageBody());
                    ipRemoto=sdpMessage.getIpAddress();
                    paRemoto=sdpMessage.getMediaPort(SdpMessage.AUDIO);
                    pvRemoto=sdpMessage.getMediaPort(SdpMessage.VIDEO);

                    if (paRemoto!=0) {
                        string=sdpMessage.message().substring(sdpMessage
                            .message().indexOf("m=audio"));
                        string=string.substring(0,string.indexOf('\n')-1);
                        string=string.substring(string.lastIndexOf(' ')+1);
                        addSIP("                   Receiving AUDIO: "+
                               "Payload Type "+string+" by port "+paRemoto);
                    }
                    if (pvRemoto!=0) {
                        string=sdpMessage.message().substring(sdpMessage
                            .message().indexOf("m=video"));
                        string=string.substring(0,string.indexOf('\n')-1);
                        string=string.substring(string.lastIndexOf(' ')+1);
                        addSIP("                   Receiving VIDEO: "+
                               "Payload Type "+string+" by port "+pvRemoto);
                    }

                    addTimedSIP(" ====>> Request ACK ====>> "+remoto);

                } else if ((call!=null) && (codigo==603) &&
                    (callID.equals(call.getCallId()))) {
                    addTimedSIP(" <<==== Response DECLINE (603) <<==== "+remoto);
                    addTimedSIP(" ====>> Request ACK ====>> "+remoto+"\n");
                } else if ((call!=null) && (codigo>200) &&
                    (callID.equals(call.getCallId()))) {
                    addTimedSIP(" <<==== Response Rejecting ("+codigo+
                                ") <<==== "+remoto+"\n");
                } else if ((call!=null) && (callID.equals(call.getCallId()))) {
                    addTimedSIP(" <<==== Response ("+codigo+") <<==== "+remoto);
                } else {
                    addTimedSIP(" <<==== Unexpected Response ("+codigo+
                                ") <<==== "+
                                responseMessage.getHeaderData(SipHeader.To));
                }

                if ((call!=null) && (callID.equals(call.getCallId()))) {
                    respuestaRemota=codigo;
                }
            }

            public synchronized void requestReceived(CallEvent ce) {
                SipRequestMessage requestMessage=
                    (SipRequestMessage) ce.getMessage();
                String callID=requestMessage.getHeaderData(SipHeader.Call_ID);
                String from=requestMessage.getHeaderData(SipHeader.From);
                String string=""; //cadena temporal para obtener Payload Types

                if (requestMessage.getMethod()==Sip.INVITE) { //Recibido INVITE
                    SipCall tempCall; //variable temporal con llamada entrante
                    SipCallMember tempMember; //variable temporal con miembro

                    if ((call!=null) || (solicitaLlamada) || (solicitaColgar)) {
                        //No se desean aceptar solicitudes INVITE

                        //Capturamos la llamada entrante indeseada y rechazamos
                        for (java.util.ListIterator li=client.getCallList()
                                .listIterator(); li.hasNext(); ) {

                            tempCall=(SipCall) li.next();

                            if (callID.equals(tempCall.getCallId())) {
                                tempMember=tempCall.getMember(new SipUri(from));

                                addTimedSIP(" <<==== Request INVITE <<==== "+
                                            tempMember.getUri().uri());
                                addTimedSIP(" ====>> Response TRYING (100) "+
                                            "====>> "+tempMember.getUri().uri());
                                addTimedSIP(" ====>> Response RINGING (180) "+
                                            "====>> "+tempMember.getUri().uri());
                                tempMember.declineInvite(requestMessage.message(),
                                                         TIPO_SDP);
                                addTimedSIP(" ====>> Response DECLINE (603) "+
                                            "====>> "+tempMember.getUri().uri());
                                rejectingCall=tempCall;

                            }
                        }

                    } else { //Se aceptan INVITE

                        //Capturamos la nueva llamada entrante
                        for (java.util.ListIterator li=client.getCallList()
                                .listIterator(); li.hasNext(); ) {

                            tempCall=(SipCall) li.next();

                            if (callID.equals(tempCall.getCallId())) {
                                call=tempCall; //Llamada encontrada
                            }
                        }

                        if (call==null) {
                            addTimedSIP(" <<==== Received a wrong INVITE signal");
                        } else {
                            member=call.getMember(new SipUri(from));
                            SdpMessage sdpMessage=new SdpMessage(requestMessage
                                                                .messageBody());

                            remoto=member.getUri().uri();
                            ipRemoto=sdpMessage.getIpAddress();
                            paRemoto=sdpMessage.getMediaPort(SdpMessage.AUDIO);
                            pvRemoto=sdpMessage.getMediaPort(SdpMessage.VIDEO);
                            paLocal=(paRemoto==0)?AUDIO_PORT:paRemoto+10;
                            pvLocal=(pvRemoto==0)?VIDEO_PORT:pvRemoto+10;

                            addTimedSIP(" <<==== Request INVITE <<==== "+remoto);
                            if (paRemoto!=0) {
                                string=sdpMessage.message().substring(sdpMessage
                                    .message().indexOf("m=audio"));
                                string=string.substring(0,string.indexOf('\n')-1);
                                string=string.substring(string.lastIndexOf(' ')+1);
                                addSIP("                   Receiving AUDIO: "+
                                       "Payload Type "+string+" by port "+paRemoto);
                            }
                            if (pvRemoto!=0) {
                                string=sdpMessage.message().substring(sdpMessage
                                    .message().indexOf("m=video"));
                                string=string.substring(0,string.indexOf('\n')-1);
                                string=string.substring(string.lastIndexOf(' ')+1);
                                addSIP("                   Receiving VIDEO: "+
                                       "Payload Type "+string+" by port "+pvRemoto);
                            }

                            addTimedSIP(" ====>> Response TRYING (100) ====>> "
                                        +remoto);
                            addTimedSIP(" ====>> Response RINGING (180) ====>> "
                                        +remoto);
                            solicitaLlamada=true;
                        }
                    }

                } else if (requestMessage.getMethod()==Sip.ACK) { //Recibido ACK
                    if ((rejectingCall!=null) &&
                        (callID.equals(rejectingCall.getCallId()))) {

                        client.deleteCall(rejectingCall);
                        rejectingCall=null;
                        addTimedSIP(" <<==== Request ACK <<==== "+
                                    from+"\n");
                    } else if ((call!=null) && (callID.equals(call.getCallId()))) {
                        addTimedSIP(" <<==== Request ACK <<==== "+remoto);
                    } else {
                        addTimedSIP(" <<==== Unexpected Request ACK <<==== "+
                                    from+"\n");
                    }

                } else if (requestMessage.getMethod()==Sip.CANCEL) { //CANCEL
                    if (((call!=null) && callID.equals(call.getCallId()))) {
                        addTimedSIP(" <<==== Request CANCEL <<==== "+remoto);
                        addTimedSIP(" ====>> Response OK (200) ====>> "+remoto+
                                    "\n");
                        member.acceptInfo("",TIPO_SDP);
                        member=null;
                        if (call!=null) {
                            client.deleteCall(call);
                            call=null;
                        }
                        respuestaRemota=0;
                        solicitaColgar=true;
                        solicitaLlamada=false;
                    } else {
                        addTimedSIP(" <<==== Unexpected Request CANCEL <<==== "+
                                    from);
                        addTimedSIP(" ====>> Response NOT EXIST (481) ====>> "+
                                    from+"\n");
                    }

                } else if (requestMessage.getMethod()==Sip.BYE) { //Recibido BYE
                    if (((call!=null) && callID.equals(call.getCallId()))) {
                        addTimedSIP(" <<==== Request BYE <<==== "+remoto);
                        addTimedSIP(" ====>> Response OK (200) ====>> "+remoto+
                                    "\n");
                        member.acceptBye("",TIPO_SDP);
                        member=null;
                        if (call!=null) {
                            client.deleteCall(call);
                            call=null;
                        }
                        respuestaRemota=0;
                        solicitaColgar=true;
                    } else {
                        addTimedSIP(" <<==== Unexpected Request BYE <<==== "+
                                    from);
                        addTimedSIP(" ====>> Response NOT EXIST (481) ====>> "+
                                    from+"\n");
                    }


                } else if (requestMessage.getMethod()==Sip.MESSAGE) { //MESSAGE
                    String message=requestMessage.messageBody();
                    addTimedSIP(" <<==== Request MESSAGE <<==== "+from);
                    addSIP("                   Message received is: ["+message+
                           "]\n");
                }
            }

        };
        client.addCallListener(cl);

        addSIP("\nClient Sip ready with user: "+usuario+"@"+ipLocal+"\n");

    }

    /**
     * Metodo encargado de enviar la peticion SIP INVITE
     * @param uriRemoto SIP URI al que se quiere realizar la peticion INVITE
     */
    public void llamar(String uriRemoto) {

        remoto=uriRemoto;
        if (remoto.indexOf('@')<0) remoto="anonymous@"+remoto;

        if (!(new SipUri(remoto)).isValid()) {
            respuestaRemota=485; //Error por URI ambigua
        } else {

            //SIP URI correcta. Enviar SIP INVITE
            SdpMessage sdpMessage=new SdpMessage();
            sdpMessage.setName("jSIP");
            sdpMessage.setUsername(localUser.toString());
            sdpMessage.setIpAddress(ipLocal);
            sdpMessage.setPort(AUDIO_PORT);
            if (audioPT!=null) {
                paLocal=AUDIO_PORT;
                sdpMessage.addMedia(SdpMessage.AUDIO,audioPT,AUDIO_TRANSPORT,
                                    paLocal);
            } else paLocal=0;
            if (videoPT!=null) {
                pvLocal=VIDEO_PORT;
                sdpMessage.addMedia(SdpMessage.VIDEO,videoPT,VIDEO_TRANSPORT,
                                    pvLocal);
            } else pvLocal=0;

            call=new SipCall(localUser, null); //Creamos una llamada call nueva
            client.addCall(call);

            member=new SipCallMember(call,new SipUri(remoto));
            member.requestInvite(sdpMessage.message(),TIPO_SDP);
            addTimedSIP(" ====>> Request INVITE ====>> "+remoto);
            if (audioPT!=null) addSIP("                   Transmitting AUDIO: "+
                                      "Payload Type "+audioPT+
                                      " by port "+paLocal);
            if (videoPT!=null) addSIP("                   Transmitting VIDEO: "+
                                      "Payload Type "+videoPT+
                                      " by port "+pvLocal);

            respuestaRemota=0;
        }
    }

    /**
     * Metodo encargado de cerrar la conexion
     */
    public void colgar() {
        if (member==null) {
            //Miembro ya desconectado
            if (call!=null) {
                client.deleteCall(call);
                call=null;
            }
        } else {
            member.localStatusUpdated();
            if (member.getStatus()==member.Disconnected) {
                //Miembro ya desconectado
                member=null;
                if (call!=null) {
                    client.deleteCall(call);
                    call=null;
                }
            } else if (member.getStatus()<member.Connected) {
                member.requestCancel(); //Si no se ha conectado todava,cancelar
                addTimedSIP(" ====>> Request CANCEL ====>> "+remoto);
                if (call!=null) {
                    closingCall=call;
                    call=null;
                }
            } else {
                member.requestDisconnect("",TIPO_SDP); //Solicitar Desconexion
                addTimedSIP(" ====>> Request BYE ====>> "+remoto);
                if (call!=null) {
                    closingCall=call;
                    call=null;
                }
            }
            member=null;
        }
        respuestaRemota=0;
	solicitaColgar=false;
    }

    /**
     * Metodo para indicar que se ha aceptado la llamada solicitada
     */
    public void aceptar() {
        if (solicitaLlamada==true) {

            SdpMessage sdpMessage=new SdpMessage();
            sdpMessage.setName("jSIP");
            sdpMessage.setUsername(localUser.toString());
            sdpMessage.setIpAddress(ipLocal);
            sdpMessage.setPort(AUDIO_PORT);
            if (audioPT!=null) {
                sdpMessage.addMedia(SdpMessage.AUDIO,audioPT,AUDIO_TRANSPORT,
                                    paLocal);
            } else paLocal=0;
            if (videoPT!=null) {
                sdpMessage.addMedia(SdpMessage.VIDEO,videoPT,VIDEO_TRANSPORT,
                                    pvLocal);
            } else pvLocal=0;

            client.addCall(call);

            member.acceptInvite(sdpMessage.message(),TIPO_SDP);
            addTimedSIP(" ====>> Response OK (200) ====>> "+remoto);
            if (audioPT!=null) addSIP("                   Transmitting AUDIO: "+
                                      "Payload Type "+audioPT+
                                      " by port "+paLocal);
            if (videoPT!=null) addSIP("                   Transmitting VIDEO: "+
                                      "Payload Type "+videoPT+
                                      " by port "+pvLocal);


            solicitaLlamada=false;
        }
    }

    /**
     * Metodo para indicar que se ha rechazado la llamada solicitada
     */
    public void rechazar() {
        if (solicitaLlamada==true) {
            if (member!=null) {
                member.declineInvite("",TIPO_SDP);
                addTimedSIP(" ====>> Response DECLINE (603) ====>> "+remoto);
            }
            if (call!=null) {
                rejectingCall=call;
                call=null;
            }

            respuestaRemota=0;
            solicitaLlamada=false;
        }
    }

    /**
    * Metodo que establece la informacion de depuracion sobre el protocolo SIP
    * @param string Cadena de texto con la informacion de depuracion sobre SIP
    */
    public synchronized void setSIP(String string) {
        sipDebug=new String(string);
    }

    //Metodo que aade informacion con el tiempo en el que se produjo
    private synchronized void addTimedSIP(String string) {
        long t=System.currentTimeMillis(); //Hora en milisegundos
        String time="";

        t/=1000; //Convertimos a segundos
        time=(t%60)+time;
        if ((t%60)<10) time="0"+time;
        t/=60; //Convertimos a minutos
        time=(t%60)+":"+time;
        if ((t%60)<10) time="0"+time;
        t/=60; //Convertimos a horas
        time=(t%24)+":"+time;
        if ((t%24)<10) time="0"+time;
        sipDebug=new String(sipDebug+time+" | "+
                            client.getContactUri().uri()+string+"\n");
    }

    /**
    * Metodo que aade informacion adicional de depuracion a la existente
    * @param string Cadena de texto con la nueva informacion a aadir
    */
    public synchronized void addSIP(String string) {
        sipDebug=new String(sipDebug+string+"\n");
    }


    /**
     * Metodo que obtiene la informacion de depuracion sobre el protocolo SIP
     * @return String Cadena de texto con la informacion de depuracion sobre SIP
     */
    public synchronized String getSIP() {
        return sipDebug;
    }

    /**
     * Destructor del Agente de Usuario. Se libera la memoria ocupada.
     */
    public void salir() {
        if (solicitaLlamada==true) {
            member.declineInvite("",TIPO_SDP);
        }
        if (call!=null) colgar();
        client.removeCallListener(client.getCallListener());

        try {finalize();} catch (Throwable e) {}
    }

}
