/*
*   umasip is a Videoconference program that uses the SIP protocol (RFC 3261)
*   Copyright (C) 2004 Pablo Montoro Escao
*
*   This program is free software; you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation; either version 2 of the License, or
*   (at your option) any later version.
*
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with this program; if not, write to the Free Software
*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*/
import java.io.*;

/**
 * <p>Programa: umasip</p>
 * <p>Copyright: Copyright GNU/GPL (c) 2004 Pablo Montoro Escao</p>
 * @author Pablo Montoro Escao
 * @version 1.00
 * @since SDK 1.4.2.
 */

public class Idioma {

    public final int VERSION=umasip.VERSION;
    public final String DEFAULT="Espaol"; //Idioma seleccionado por Defecto
    private final String NOT_FOUND="---"; //Si falla la carga de texto

    public boolean ready=false; //Indica si esta cargado y preparado el idioma
    private RandomAccessFile fichero;
    private long punteroInicio;

    public Disponibles disponibles; //Idiomas disponibles, nombre y file

    public class Disponibles {
        public String[] nombre;
        public File[] file;

        public Disponibles() {
            File[] lista=(new File("lang")).listFiles();
            int index=0; //Indice de los arrays previos de nombre y file
            String[] preNombre=new String[lista.length];
            File[] preFile=new File[lista.length];
            int numero;
            String cadena;

            for (int i=0; i<lista.length; i++) {
                if (lista[i].getName().endsWith(".fiu") && lista[i].canRead())
                    try {
                        fichero=new RandomAccessFile(lista[i],"r");
                        if ( (leerInt(fichero)<=VERSION) &&
                            (fichero.readLine().equals("Fichero de Idioma")) ) {

                           do {
                               numero=leerInt(fichero);
                               cadena=fichero.readLine();
                           } while (numero!=0);

                           preNombre[index]=new String(cadena);
                           preFile[index++]=lista[i];
                        }
                    } catch (Exception e) {}
            }

            //Construimos los campos nombre y file de los idiomas disponibles
            nombre=new String[index];
            file=new File[index];
            for (int i=0; i<index; i++) {
                nombre[i]=new String(preNombre[i]);
                file[i]=preFile[i];
            }
        }

        public boolean isDisponible(int idioma) {
            boolean ok=false;
            if ((idioma>=0) && (idioma<file.length)) {
                if (file[idioma].canRead()) ok=true;
            }
            return ok;
        }

        public void salir() {
            try {finalize();} catch (Throwable e) {}
        }
    }

    /**
     * Constructor del modulo de idioma. Se toma el idioma por defecto.
     */
    public Idioma() {
        init(DEFAULT); //Idioma por defecto
    }

    /**
     * Constructor del modulo de idioma.
     * @param idioma Indica el nombre del idioma
     */
    public Idioma(String idioma) {
        init(idioma);
    }

    //Metodo constructor general
    private void init(String idioma) {
        disponibles=new Disponibles();
        if (!cargar(idioma)) if (!cargar(DEFAULT)) cargar(0);
        //Si no carga idioma cargar DEFAULT y si no cargar el primero (0)
    }

    //Lee un numero entero positivo hasta encontrar un caracter no numerico
    //El caracter no numerico tambien lo lee, quitandolo del buffer
    //En caso de no existir un caracter numerico devuelve el valor -1 (error)
    private int leerInt(RandomAccessFile fichero) throws IOException {
        int numero;
        int entero;

        numero=fichero.readByte()-'0'; //Leemos el primer byte y su valor
        if ((numero<0)||(numero>9)) entero=-1; //Devuelve valor erroneo
        else entero=0;
        while ((numero>=0)&&(numero<=9)) {
            entero=10*entero+numero;
            numero=fichero.readByte()-'0';
        }
        return entero;
    }

    /**
     * Carga el fichero de idioma indicado.
     * @param idioma Indica el nombre del idioma
     * @return si se ha realizado la carga del idioma correctamente o no
     */
    public boolean cargar(String idioma) {
        int indiceIdioma=-1;

        for (int i=0; i<disponibles.nombre.length; i++) {
            if (disponibles.nombre[i].toUpperCase()
                .equals(idioma.toUpperCase())) indiceIdioma=i;
        }
        return cargar(indiceIdioma);
    }

    /**
     * Carga el fichero de idioma indicado.
     * @param idioma Indica el indice del idioma
     * @return si se ha realizado la carga del idioma correctamente o no
     */
    public synchronized boolean cargar(int idioma) {
        if (ready==true) {
            ready=false;
            try {fichero.close();} catch (Exception e) {}
        }

        if (disponibles.isDisponible(idioma)) try {
            fichero=new RandomAccessFile(disponibles.file[idioma],"r");
            if ((leerInt(fichero)<=VERSION) &&
                (fichero.readLine().equals("Fichero de Idioma"))) {

                punteroInicio=fichero.getFilePointer();
                ready=true;
            } else {
                try {fichero.close();} catch (Exception e) {}
            }
        } catch (Exception e) {}

        return ready;
    }

    /**
     * Devuelve la linea de texto indicada en el idioma cargado.
     * @param linea el numero correspondiente al texto a recuperar
     * @return el texto correspondiente a la linea de texto solicitada
     */
    public synchronized String texto(int linea) {
        int numero;
        String cadena;

        if (!ready) cadena=NOT_FOUND;
        else {
            try {
                fichero.seek(punteroInicio);
                do {
                    numero=leerInt(fichero);
                    cadena=fichero.readLine();
                } while (numero!=linea);
            } catch (Exception e) {
                cadena=NOT_FOUND;
            }
        }
        return cadena;
    }

    /**
     * Destructor del Modulo de idioma.
     * Se libera la memoria ocupada y se cierran los ficheros de idioma.
     */
    public void salir() {
        ready=false;
        try {fichero.close();} catch (Exception e) {}
        disponibles.salir();
        try {finalize();} catch (Throwable e) {}
    }

}
