/*
 * This file was derived from libdissipate, an open source SIP library. The original file
 * was modified on 1/23/2001.  Please see
 * http://www.div8.net/dissipate for more information.
 *
 * Copyright (c) 2000 Billy Biggs <bbiggs@div8.net>
 *
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to the
     * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 */
/**
 * class SdpMessage
 *
 * Class for representing an SDP message.
 *
 * This code has been generated using C2J++
 * C2J++ is based on Chris Laffra's C2J (laffra@watson.ibm.com)
 * Read general disclaimer distributed with C2J++ before using this code
 * For information about C2J++, send mail to Ilya_Tilevich@ibi.com
 *
 */
package org.mitre.jsip;

import java.util.*;

public class SdpMessage {

    /**
     * Construct a blank SDP message object.
     */
    public SdpMessage() {
        valid=true;
        port=0;
        payload_types=new int[1];
        payload_types[0]=0; // set gsm as a default
        initMediaLists();
    }

    /**
     * Parse the given SDP message.
     *
     * @param parseinput
     */
    public SdpMessage(String parseinput) {
        initMediaLists();
        parseInput(parseinput);
    }

    private void initMediaLists() {
        // init all the media types
        audioML=new MediaList("audio");
        videoML=new MediaList("video");
        dataML=new MediaList("data");
        applicationML=new MediaList("application");
        controlML=new MediaList("control");
    }

    /**
     * Returns true if the parsed SDP message is valid.
     *
     * @return bool
     */
    public boolean isValid() {
        return valid;
    }

    /**
     * Returns true if this SDP implies the call is on hold.
     *
     * @return bool
     */
    public boolean isOnHold() {
        return (ipaddress.trim()=="0.0.0.0");
    }

    /**
     * Returns the unique session name.  This is the s= line of SDP.
     *
     * @return String
     */
    public String getName() {
        return sessionname;
    }

    /**
     * Returns the IP address of the session described.
     *
     * @return String
     */
    public String getIpAddress() {
        return ipaddress;
    }

    /**
     * Returns the port on which media is being received.
     *
     * @return int
     */
    public int getPort() {
        return port;
    }

    /**
     * Returns the port on which the media type specified is being received.
     *
     * @param mediaType Type of media which port is to be returned
     * @return int
     */
    public int getMediaPort(int mediaType) {
        if (mediaType==AUDIO) return audioML.getPort();
        else if (mediaType==VIDEO) return videoML.getPort();
        else if (mediaType==DATA) return dataML.getPort();
        else if (mediaType==APPLICATION) return applicationML.getPort();
        else if (mediaType==CONTROL) return controlML.getPort();
        else return 0;
    }

    /**
     * Sets the IP address to which media should be directed.
     *
     * @param newaddr
     */
    public void setIpAddress(String newaddr) {
        ipaddress=newaddr;
    }

    /**
     * Sets the port on which media is to be received.
     *
     * @param newport
     */
    public void setPort(int newport) {
        port=newport;
    }

    /**
     * Sets the name of the session (the SDP s= line).
     *
     * @param newname
     */
    public void setName(String newname) {
        sessionname=newname;
    }

    /**
     * Sets the username (for the SDP o= line).
     *
     * @param username
     */
    public void setUsername(String username) {
        this.username=username;
    }

    /**
     * Sets the times the session is active
     */
    public void setSessionTimes(Date startTime, Date stopTime) {
        this.startTime=startTime;
        this.stopTime=stopTime;
    }

    /**
     * Allows the addition of a standard protocol
     */
    public void addMedia(int media, String type, String protocol,
                         int port) {
        switch (media) {
            case AUDIO:
                audioML.add(type);
                audioML.setTransportProtocol(protocol);
                audioML.setPort(port);
                break;
            case VIDEO:
                videoML.add(type);
                videoML.setTransportProtocol(protocol);
                videoML.setPort(port);
                break;
            case APPLICATION:
                applicationML.add(type);
            case DATA:
                applicationML.add(type);
            case CONTROL:
                applicationML.add(type);
        }
    }

    public void addMediaAttribute(int media, String type,
                                  String attStr) {
        switch (media) {
            case AUDIO:
                audioML.addAttribute(type, attStr);
                break;
            case VIDEO:
                videoML.addAttribute(type, attStr);
                break;
            case APPLICATION:
                applicationML.addAttribute(type, attStr);
                break;
            case DATA:
                applicationML.addAttribute(type, attStr);
                break;
            case CONTROL:
                applicationML.addAttribute(type, attStr);
                break;
        }
    }

    /**
     * Sets the payload types (for the SDP m= line)
     *
     * @param types The payload types the application can use
     *
     public void setPayloadTypes(int[] types)
     {
     payload_types = new int[types.length];
     System.arraycopy(types, 0, payload_types, 0, types.length);
     }
     */

    /**
     * Parses the given input as an SDP message.
     *
     * @param parseinput
     */
    private void parseInput(String parseinput) {
        String input;
        String curline;
        char lineindex;
        String portstr;
        String codecstr;
        valid=false;

        port=0;
        ipaddress=null;

        StringTokenizer lineTok=new StringTokenizer(parseinput, "\n");
        String verString=lineTok.nextToken().trim();

        if (!verString.equals("v=0")) {
            System.err.println("Bad SDP message. Returning");
            return;
        }

        while (lineTok.hasMoreTokens()) {
            curline=lineTok.nextToken().trim();

            if (curline.length()==0) lineindex=' ';
            else lineindex=curline.charAt(0);

            if (lineindex=='c') {
                ipaddress=curline.substring(curline.indexOf("IP4")+4).
                          trim();
            }
            else if (lineindex=='m') {
                StringTokenizer tok=new StringTokenizer(curline);
                String media=tok.nextToken().substring(2);
                portstr=tok.nextToken();
                int mport;
                try {
                    mport=Integer.parseInt(portstr);
                }
                catch (NumberFormatException nfe) {
                    mport=0;
                }

                String transstr=tok.nextToken();

                // any more tokens are the list of formats allowed
                while (tok.hasMoreTokens()) {
                    String newType=tok.nextToken();
                    addMedia(getMediaType(media), newType,
                             transstr, mport);
                }

            }
            else if (lineindex=='s') {
                setName(curline.substring(curline.indexOf("=")+1));
            }
            else if (lineindex=='t') {
                String timestr=curline.substring(curline.indexOf("=")+
                                                 1);
                // TODO something with this
            }
            else if (lineindex=='o') {
                StringTokenizer oTok=new StringTokenizer(curline);
                String username=oTok.nextToken().substring(2);
                setUsername(username);
                String sessid=oTok.nextToken();
                String version=oTok.nextToken();
                String nwtype=oTok.nextToken();
                String addtype=oTok.nextToken();
                String originator=oTok.nextToken();
            }

        }

        valid=true;
    }

    /**
     * Builds and returns an SDP message given the contents.
     *
     * @return String
     */
    public String message() {
        Date date=new Date(); // to make a timestamp.

        StringBuffer buffer=new StringBuffer();

        // Version
        buffer.append("v=0\r\n");

        String idVer=String.valueOf(date.getTime())+" ";

        // Session Id
        buffer.append("o="+username+" "+idVer+idVer+"IN IP4 "
                      +SipUtil.getLocalFqdn()+"\r\n");

        // Session Name
        buffer.append("s="+sessionname+"\r\n");

        // Connection Info
        buffer.append("c=IN IP4 "+ipaddress+"\r\n");

        // Time active
        if (startTime!=null) {
            long srttime=getNTPTime(startTime);
            long stptime=getNTPTime(stopTime);
            buffer.append("t="+Long.toString(srttime)+" "+
                          Long.toString(stptime)+"\r\n");
        }
        else {
            buffer.append("t=0 0\r\n");
        }

        // Media information
        if (audioML.size()>0) {
            buffer.append(audioML.printMLine());
        }
        if (videoML.size()>0) {
            buffer.append(videoML.printMLine());
        }
        if (applicationML.size()>0) {
            buffer.append(applicationML.printMLine());
        }
        if (dataML.size()>0) {
            buffer.append(dataML.printMLine());
        }
        if (controlML.size()>0) {
            buffer.append(controlML.printMLine());

            // now attribute lines
        }
        if (audioML.size()>0) {
            buffer.append(audioML.printALines());
        }
        if (videoML.size()>0) {
            buffer.append(videoML.printALines());
        }
        if (applicationML.size()>0) {
            buffer.append(applicationML.printALines());
        }
        if (dataML.size()>0) {
            buffer.append(dataML.printALines());
        }
        if (controlML.size()>0) {
            buffer.append(controlML.printALines());
            //	msg += "a=ptime:20\r\n";

        }
        return buffer.toString();
    }

    private String createPayloadString() {
        String ret="";

        for (int i=0; i<payload_types.length; i++) {
            ret+=String.valueOf(payload_types[i])+" ";

        }
        return ret;
    }

    private static final long NTP_UNIX_DIFFERENCE=2208988800L;

    private long getNTPTime(Date ourTime) {
        // according to RFC 2327 (not really the definitive source) NTP is
        // ahead of unix time by the above amount. NTP time is based off of
        // the clock starting at 1 Jan 1900 instead of 1970.
        return (ourTime.getTime()/1000)+NTP_UNIX_DIFFERENCE;
    }

    private int getMediaType(String mediaStr) {
        if (mediaStr==null) {
            return-1;
        }

        if (mediaStr.equals("audio")) {
            return AUDIO;
        }
        if (mediaStr.equals("video")) {
            return VIDEO;
        }
        if (mediaStr.equals("application")) {
            return APPLICATION;
        }
        if (mediaStr.equals("data")) {
            return DATA;
        }
        if (mediaStr.equals("control")) {
            return CONTROL;
        }

        return-1;
    }

    /**
     * test out SdpMessage
     */

    public static void main(String[] args) {
        String mcastAddr="225.5.5.5/15";
        int mcastPort=2556;

        // create a message from scratch and print it out
        SdpMessage sdpMessage=new SdpMessage();
        sdpMessage.setName("jSIP");
        sdpMessage.setUsername("steve");
        sdpMessage.addMedia(SdpMessage.AUDIO, "0", "RTP/AVP",
                            mcastPort);
        sdpMessage.addMediaAttribute(SdpMessage.AUDIO, "0",
                                     "a=ptime:30");
        sdpMessage.addMedia(SdpMessage.AUDIO, "3", "RTP/AVP",
                            mcastPort);
        sdpMessage.setIpAddress(mcastAddr);
        sdpMessage.setPort(mcastPort);

        System.out.println("Test message is\n\n"+sdpMessage.message());

        System.out.println("\n\nNow parsing same message...\n");
        SdpMessage sdpMessage2=new SdpMessage(sdpMessage.message());
        System.out.println("Second message is:\n"+sdpMessage2.message());
    }

    // class variables
    private boolean valid;
    private String sessionname;
    private String ipaddress;
    private String username="Unknown";
    private int port;
    private int[] payload_types;
    private Date startTime;
    private Date stopTime;

    // Possible media types
    public static final int AUDIO=0;
    public static final int VIDEO=1;
    public static final int DATA=2;
    public static final int APPLICATION=3;
    public static final int CONTROL=4;

    // Media lists. For now, just have one for each type
    MediaList audioML;
    MediaList videoML;
    MediaList dataML;
    MediaList applicationML;
    MediaList controlML;
}

/**
 * Quick implementation of media and attribute listings for SDP
 */

class MediaList
    extends Vector {
    // standard payload types as defined in RFC 1890
    private final String[][] standardPayloads= { {"0", "PCMU", "8000",
        "A", "1"}
        , {"1", "1016", "8000", "A", "1"}
        , {"2", "G721", "8000", "A", "1"}
        , {"3", "GSM", "8000", "A", "1"}
        , {"5", "DVI4", "8000", "A", "1"}
        , {"6", "DVI4", "16000", "A", "1"}
        , {"7", "LPC", "8000", "A", "1"}
        , {"8", "PCMA", "8000", "A", "1"}
        , {"9", "G722", "8000", "A", "1"}
        , {"10", "L16", "44100", "A", "2"}
        , {"11", "L16", "44100", "A", "1"}
        , {"15", "G728", "8000", "A", "1"}
        , {"25", "CelB", "90000", "V", "0"}
        , {"26", "JPEG", "90000", "V", "0"}
        , {"28", "nv", "90000", "V", "0"}
        , {"31", "H261", "90000", "V", "0"}
        , {"32", "MPV", "90000", "V", "0"}
        , {"33", "MP2T", "90000", "V", "0"}
    };

    private final int DYNAMIC_MIN=96;
    private final int DYNAMIC_MAX=127;
    private final int UDP_PORT_MIN=1024;
    private final int UDP_PORT_MAX=65535;

    private String type;
    private String transportProtocol;
    private int port;

    Vector attributes;

    MediaList(String type) {
        this(type, "RTP/AVP", 0);
    }

    MediaList(String type, String transportProtocol) {
        this(type, transportProtocol, 0);
    }

    MediaList(String type, String transportProtocol, int port) {
        attributes=new Vector();
        this.type=type;
        this.transportProtocol=transportProtocol;
        this.port=port;
    }

    public void setTransportProtocol(String transportProtocol) {
        this.transportProtocol=transportProtocol;
    }

    public void setPort(int port) {
        this.port=port;
    }

    public int getPort() {
        return this.port;
    }

    void addAttribute(String attType, String attString) {
        // check to see if type has been added. If so, allow attribute
        // to be added to the list
        for (Enumeration e=elements(); e.hasMoreElements(); ) {
            String addedType= (String) e.nextElement();
            if (addedType.equals(attType)) {
                attributes.add(attString);
            }
        }
    }

    public String printMLine() {
        // if there are no types specified, return null
        if (this.size()==0) {
            return null;
        }

        StringBuffer buf=new StringBuffer();
        buf.append("m="+type+" "+String.valueOf(port)+" "+
                   transportProtocol);

        for (int i=0; i<this.size(); i++) {
            String element= (String)this.get(i);
            buf.append(" "+element);
        }

        buf.append("\r\n");
        return buf.toString();
    }

    public boolean hasAttributes() {
        return (attributes.size()>0);
    }

    // override Vector add for standard types
    public boolean add(Object obj) {
        if (obj==null) {
            return false;
        }

        String newType= (String) obj;
        super.add(newType);

        for (int i=0; i<standardPayloads.length; i++) {
            String nType=standardPayloads[i][0]; // get the number
            if (nType.equals(newType)) {
                String typeAtt="a=rtpmap:"+newType+
                               " "+standardPayloads[i][1]+"/"+
                               standardPayloads[i][2];
                attributes.add(typeAtt);
                return true;
            }
        }
        return true;

    }

    public String printALines() {
        StringBuffer buffer=new StringBuffer();

        for (int i=0; i<attributes.size(); i++) {
            String line= (String) attributes.get(i);
            buffer.append(line+"\r\n");
        }

        return buffer.toString();
    }
}
