/*
 * This file was derived from libdissipate, an open source SIP library. The original file
 * was modified on 1/23/2001.  Please see
 * http://www.div8.net/dissipate for more information.
 *
 * Copyright (c) 2000 Billy Biggs <bbiggs@div8.net>
 *
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to the
     * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 */
/**
 * class SipCallMember
 *
 * Object to reference a member of a call.  Contains the remote URI, their
 * current Contact URI, and their current status in the call.  It also contains
 * their current session description, and a copy of our local session
 * description for this call leg.
 *
 * This class also provides an API for modifying call state to this member of
 * the call, such as sending an invitation to join the session, or accepting a
 * call.
 */
package org.mitre.jsip;

import java.awt.Toolkit;
import org.mitre.jsip.event.*;

public class SipCallMember
    extends Object {
    /**
     * Creates a new member for this call with the given URI.  Initially,
     * this sets the Contact URI for the call member to be the same.
     * @param parent
     * @param uri
     */
    public SipCallMember(SipCall parent, SipUri uri) {
        call=parent;
        memberuri=uri;
        contacturi=uri;
        lclTrans=null;
        rmtTrans=null;
        status=Disconnected;
        parent.addMember(this);
    }

    /**
     * Returns the current status of the call member.
     *
     * @return Status
     */
    public int getStatus() {
        return status;
    }

    /**
         * Returns the most recent session description provided by the member
     * for sending media to them.
     *
     * @return String
     */
    public String getSessionDescription() {
        return sessiondesc;
    }

    /**
         * Returns the MIME Content Type of the session description provided by
         * the call member.  If none was provided, this function returns NULL.
     * @return MimeContentType
     */
    public MimeContentType getSessionDescriptionType() {
        return sessiontype;
    }

    /**
         * Returns the most recently sent local session description.  Provided
     * for reference.
     * @return String
     */
    public String getLocalSessionDescription() {
        return localsessiondesc;
    }

    /**
     * Returns the MIME Content Type of the most recently sent local
     * session description.
     * @return MimeContentType
     */
    public MimeContentType getLocalSessionDescriptionType() {
        return localsessiontype;
    }

    /**
     * Returns a text description of our current status.  Basically, this
     * is the text from the response line of the last message we received,
     * or a text description of what we're currently doing or waiting for.
     * Useful for showing the user what is going on.
     * @return String
     */
    public String getLocalStatusDescription() {
        return statusdesc;
    }

    /**
     * Returns the most recent message body we received that was not a
     * session description.
     * @return String
     */
    public String getMostRecentMessageBody() {
        return recentbody;
    }

    /**
     * Returns the MIME type of the most recent message body we received
     * that was not a session description.
     * @return MimeContentType
     */
    public MimeContentType getMostRecentMessageBodyType() {
        return recentbodytype;
    }

    /**
     * Sends a SIP INVITE request, asking the member to join in the session
     * described in the given body.  The body and MIME type provided will
     * become the new local session description for this call member.
     */

    public void requestInvite(String body, MimeContentType bodytype) {
        if (body!=null) {
            localsessiondesc=body;
            localsessiontype=bodytype;
        }

        if (status==Connected) {
            status=ReInviteRequested;
            statusdesc="Requesting session update";
        }
        else {
            status=InviteRequested;
            statusdesc="Requesting remote end to join session";
        }

        lclTrans=call.newRequest(this, Sip.INVITE, body, bodytype);
        // *** LOOK AT ***
        //connect( local, statusUpdated(), this, localStatusUpdated() );

        //statusUpdated();
        localStatusUpdated();
    }

    /**
     * Send a CANCEL msg for the given request
     */
    public void requestCancel() {
        if (status==Cancelling) {
            return;
        }

        status=Cancelling;
        statusdesc="Cancelling";
        lclTrans.cancelRequest();
    }

    /**
     * Sends a SIP BYE request to disconnect the session.
     */

    public void requestDisconnect(String body,
                                  MimeContentType bodytype) {
        if (status==Disconnected) {
            return;
        }

        status=Disconnecting;
        statusdesc="Disconnecting";
        lclTrans=call.newRequest(this, Sip.BYE, body, bodytype);
        // *** LOOK AT ***
        //connect ( local, statusUpdate(), this, localStatusUpdated() );

        //statusUpdated();
        //localStatusUpdated();
    }

    /**
         * Disconnects the session with a request to transfer to another party.
     */
    public void requestTransfer(SipUri transferto, String body,
                                MimeContentType bodytype) {
        if (status==Disconnected) {
            return;
        }

        status=Disconnecting;
        statusdesc="Transfering";
        lclTrans=call.newRequest(this, Sip.BYE, body, bodytype,
                                 transferto);
        // **** LOOK AT ****
        //connect( local, statusUpdated(), this, localStatusUpdated() );

        //statusUpdated();
        //localStatusUpdated();
    }

    /**
     * Sends a MESSAGE request
     */
    public void requestMessage(String messageBody) {

        statusdesc="Sending message";
        status=MessageRequested;
        MimeContentType bodytype=new MimeContentType("text/plain");

        lclTrans=call.newRequest(this, Sip.MESSAGE, messageBody,
                                 bodytype);

        //	localStatusUpdated();
    }

    /**
         * Sends a SIP OPTIONS request, asking the member what they support.
         * The body and MIME type provided serve no known purpose at this time.
     * The response to the OPTIONS request will become the new remote
         * session description, so this should not be called on an active call.
     * It is provided here for consistency.
     */

    public void requestOptions(String body, MimeContentType bodytype) {

        statusdesc="Querying options";
        lclTrans=call.newRequest(this, Sip.OPTIONS, body, bodytype);
        // *** LOOK AT ***
        //connect( local, statusUpdated(), this, localStatusUpdated() );

        //statusUpdated();
        //localStatusUpdated();
    }

    /**
     * Requests an INFO message
     * @param body The body of the INFO message
     * @param bodytype The content type of the body
     */
    public void requestInfo(String body, MimeContentType bodytype) {
        lclTrans=call.newRequest(this, Sip.INFO, body, bodytype);
    }

    // **********************************************
    // *********** Presence functions ***************
    // **********************************************

    /**
     * Request a SUBSCRIBE message be sent
     *
     */
    public void requestSubscribe(String body,
                                 MimeContentType bodytype) {
        statusdesc="Subscribe sent";
        lclTrans=call.newRequest(this, Sip.SUBSCRIBE, body, bodytype);

        localStatusUpdated();
    }

    /**
     * Request a NOTIFY message be sent
     *
     */
    public void requestNotify(String body, MimeContentType bodytype) {
        statusdesc="Notify sent";
        lclTrans=call.newRequest(this, Sip.NOTIFY, body, bodytype);

        //	localStatusUpdated();
    }

    /**
         * Accepts the invitation to join the session sent by the call member.
         * The body and MIME type provided will become the new local session
     * description for this call member.
     */

    public void acceptInvite(String body, MimeContentType bodytype) {
        if (rmtTrans==null) {
            return;
        }

        if (body!=null) {
            localsessiondesc=body;
            localsessiontype=bodytype;
        }

        rmtTrans.sendResponse(new SipStatus(200), body, bodytype);
        status=Connected;
        statusdesc="Connected";

        //statusUpdated();
        localStatusUpdated();
    }

    public void acceptBye(String body, MimeContentType bodytype) {
        System.out.println("In END Call............... "+rmtTrans);

        if (rmtTrans==null) {
            return;
        }

        if (body!=null) {
            localsessiondesc=body;
            localsessiontype=bodytype;
        }

        rmtTrans.sendResponse(new SipStatus(200), body, bodytype);
        status=Disconnected;
        statusdesc="Disconnected";

        //statusUpdated();
        //localStatusUpdated();
    }

    public void acceptInfo(String body, MimeContentType bodytype) {
        /*        try
                {
                    Thread.sleep(300);
                }
                catch(InterruptedException e)
                {
                    e.printStackTrace();
                }
         */
//        System.out.println("In Accept INFO............... " + rmtTrans);

        if (rmtTrans==null) {
            return;
        }

        if (body!=null) {
            localsessiondesc=body;
            localsessiontype=bodytype;
        }
        rmtTrans.sendResponse(new SipStatus(200), body, bodytype);
        //statusUpdated();
//        localStatusUpdated();
    }

    /**
         * Declines the invitation to join the session sent by the call member.
         * The body and MIME type provided are for possibly giving a reason as
     * to why the call was declined.
     */

    public void declineInvite(String body, MimeContentType bodytype) {
        if (rmtTrans==null) {
            return;
        }

        rmtTrans.sendResponse(new SipStatus(603), body, bodytype);
        status=Disconnected;
        statusdesc="Rejecting call invitation";
        //statusUpdated();
        localStatusUpdated();
    }

    /**
     * Returns the URI for this call member.
     *
     * @return SipUri &
     */
    public SipUri getUri() {
        return memberuri;
    }

    /**
     * Returns the current Contact URI for this call member.
     *
     * @return SipUri &
     */
    public SipUri getContactUri() {
        return contacturi;
    }

    /**
     * Sets the Contact URI for this call member.
     *
     * @param newcontact
     */
    public void setContactUri(SipUri newcontact) {
        contacturi=newcontact;
    }

    /**
     * Updates the URI for this call member.
     *
     * @param newuri
     */
    public void setUri(SipUri newuri) {
        memberuri=newuri;
    }

    /**
     * Returns the list of URIs where we were redirected.
     *
     * @return SipUriList &
     */
    public SipUriList getRedirectList() {
        return redirectlist;
    }

    /**
     * This signal is sent whenever the status of this call member has changed.
     *
     */

    //private void statusUpdated() {
    // SipCall.statusUpdated();
    //}

    /**
     * localStatusUpdated
     */
    public void localStatusUpdated() {
        MimeContentType mtype;

        if (lclTrans==null) {
            if (SipClient.DEBUG) {
                System.out.println("SipCallMember: "+
                    "Received what was likely a retransmission, ignoring...");
            }
            return;
        }

        if (SipClient.DEBUG) {
            System.out.println("SipCallMember: localStatusUpdated: "+
                               lclTrans.getStatus().getCode());

        }
        mtype=lclTrans.getFinalContentType();
        if (mtype!=null) {
            if (mtype.type().equals("application/sdp")) {
                sessiondesc=lclTrans.getFinalMessageBody();
                localsessiontype=mtype;
            }
            else {
                recentbody=lclTrans.getFinalMessageBody();
                recentbodytype=mtype;
            }

            if (status==InviteRequested||status==ReInviteRequested) {
                if (lclTrans.wasCancelled()) {
                    status=Disconnected;
                    statusdesc="Request cancelled";
                    lclTrans=null;
                }
                else if (lclTrans.getStatus().getCode()<200) {
                    statusdesc=lclTrans.getStatus().getReasonPhrase();
                }
                else if (lclTrans.getStatus().getCode()<300) {
                    status=Connected;
                    statusdesc="Connected: "+
                               lclTrans.getStatus().getReasonPhrase();
                    lclTrans=null;
                }
                else if (lclTrans.getStatus().getCode()<400) {
                    status=Redirected;
                    statusdesc="Redirected: "+
                               lclTrans.getStatus().getReasonPhrase();
                    redirectlist=lclTrans.getFinalContactList();
                    lclTrans=null;
                }
                else {
                    status=Disconnected;
                    statusdesc="Call Failed: "+
                               lclTrans.getStatus().getReasonPhrase();
                    lclTrans=null;
                }
            }
            else if (status==Disconnecting) {
                if (lclTrans.getStatus().getCode()>=200) {
                    status=Disconnected;
                    statusdesc="Response: "+
                               lclTrans.getStatus().getReasonPhrase();
                    lclTrans=null;
                }
            }
            else {
                if (lclTrans.getStatus().getCode()>=200) {
                    statusdesc="Response: "+
                               lclTrans.getStatus().getReasonPhrase();
                    lclTrans=null;
                }
            }

            //statusUpdated();
        }
    }

    /**
     * remoteStatusUpdated
     */
    void remoteStatusUpdated() {
        if (status==RemoteInviteRequest) {
            if (rmtTrans.wasCancelled()) {
                status=Disconnected;
                statusdesc="Request cancelled";
                //statusUpdated();
            }
        }
    }

    /**
     * incomingTransaction
     * @param newtrans
     */
    void incomingTransaction(SipTransaction newtrans) {
        MimeContentType mtype;
        rmtTrans=newtrans;

        if (rmtTrans.getRequest().getMethod()==Sip.INVITE) {

            // *** LOOK AT ***
            //connect( remote, statusUpdated(), this, remoteStatusUpdated() );

            if (status==Disconnected) {
                status=RemoteInviteRequest;
                statusdesc="Invitation received";
                rmtTrans.sendResponse(new SipStatus(180));
            }
            else {
                status=RemoteReInviteRequest;
                statusdesc="Session update requested";
            }

            mtype=rmtTrans.getRequestMessageContentType();
            if ( (mtype!=null)&&
                (mtype.type().equals("application/sdp"))) {
                sessiontype=mtype;
                sessiondesc=rmtTrans.getRequestMessageBody();
            }
            else {
                recentbodytype=mtype;
                recentbody=rmtTrans.getRequestMessageBody();
            }

            // send a CallEvent here to notify the user of the request
            /*	  if (SipClient.DEBUG) System.out.println("Sending CallEvent request ");
             CallListener callListener = SipClient.getSipClient().getCallListener();
             if ( callListener != null ) {
             //	      Toolkit.getDefaultToolkit().getSystemEventQueue().postEvent( new CallEvent( SipClient.getSipClient(), remote.getRequest(), CallEvent.CALLREQUEST_ID ) );
             CallEvent ce = new CallEvent( newtrans, remote.getRequest(), CallEvent.CALLREQUEST_ID );
             callListener.requestReceived( ce );
             }
             */
        }
        else if (rmtTrans.getRequest().getMethod()==Sip.BYE) {

            status=Disconnected;
            statusdesc="Remote end disconnected";

        }

        // statusUpdated();
        remoteStatusUpdated(); // I think
    }

    SipUri memberuri;
    SipUri contacturi;
    int status;
    SipCall call;
    SipTransaction lclTrans;
    SipTransaction rmtTrans;
    SipUriList redirectlist;
    String sessiondesc;
    MimeContentType sessiontype;
    String localsessiondesc;
    MimeContentType localsessiontype;
    String statusdesc;
    String recentbody;
    MimeContentType recentbodytype;

    // class variables
    public static final int Disconnected=0;
    public static final int InviteRequested=1;
    public static final int ReInviteRequested=2;
    public static final int RemoteInviteRequest=3;
    public static final int RemoteReInviteRequest=4;
    public static final int Redirected=5;
    public static final int Connected=6;
    public static final int Disconnecting=7;
    public static final int MessageRequested=8;
    public static final int Cancelling=9;
}
