/*
 * This file was derived from libdissipate, an open source SIP library. The original file
 * was modified on 1/23/2001.  Please see
 * http://www.div8.net/dissipate for more information.
 *
 * Copyright (c) 2000 Billy Biggs <bbiggs@div8.net>
 *
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to the
     * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 */
/**
 * class SipMessage
 *
 * Representation of a single SIP message. SipMessage is used for creating and for parsing individual SIP
 * messages.  It also contains a retransmission timer for use when sending the
 * message.
 *
 * This code has been generated using C2J++
 * C2J++ is based on Chris Laffra's C2J (laffra@watson.ibm.com)
 * Read general disclaimer distributed with C2J++ before using this code
 * For information about C2J++, send mail to Ilya_Tilevich@ibi.com
 */
package org.mitre.jsip;

import java.util.*;
import java.net.*;

public abstract class SipMessage {
    // class variables
    public static final int Request=0;
    public static final int Response=1;
    public static final int BadType=2;

    static final String version="SIP/2.0";

    // class variables
    protected SipViaList vialist;
    protected ArrayList headerlist;
    protected SipUriList recordroute;
    protected SipUriList contactlist;
    protected int type;
    private SipStatus status;
    private SipUri requesturi;
    private String messagebody;
    private int messageSize;
    private SipUri sendToAddress;

    // constructors
    protected SipMessage() {
        setDefaultVars();
    }

    /**
     * Create a SipMessage from a buffer
     */
    public static SipMessage createMessage(String inputMessage) {
        // if message begins with "SIP", assume it's a response
        if (inputMessage.trim().startsWith("SIP")) {
            return new SipResponseMessage(inputMessage);
        }

        // if not a response, make it a request
        return SipRequestMessage.createRequestMessage(inputMessage.
            trim());
    }

    /**
     * Returns true if the message has a body.
     *
     * @return bool
     */
    public boolean haveBody() {
        return (getContentLength()>0);
    }

    /**
     * Sets the body of the SIP message.
     *
     * @param newbody
     */
    public void setBody(String newbody) {
        if (messageSize>0) {
            // double check message size
            if (messageSize>newbody.length()) {
                messageSize=newbody.length();
            }
            messagebody=newbody.substring(0, messageSize);
        }
        else {
            messagebody=newbody;
        }
        if (messagebody!=null) {
            messageSize=messagebody.length();
        }
    }

    /**
     * Generates and returns the version string for this message.
     *
     * @return String
     */
    protected String getVersionString() {
        return version;
    }

    /**
     * Returns the message body.
     *
     * @return The body of the message
     */
    public String messageBody() {
        return messagebody;
    }

    /**
     * Return the type of this message, either <code>SipMessage.Request</code> or
     * <code>SipMessage.Response</code>
     */
    public int getType() {
        return type;
    }

    /**
     * Return the Content Length of the message body
     *
     * @return The size of the message body
     */
    public int getContentLength() {
        return (messagebody==null)?0:messagebody.length();
    }

    /**
     * Gets the CSeq method type of the incoming message
     * the CSeq  header field has the CSeq and the Method of the current transaction <br>
     * This method extracts the Method from the CSeq header.  <br>
     * NOTE: This should have been in org.mitre.jsip.SipMessage
     * SRJ: does this need to be static??? are we referencing an instanciated message???
     */
    public static int getCSeqMethod(String s) {
        int intMethod;
        String strTemp=new String(s);
        strTemp=strTemp+" ";
        StringTokenizer stringToken=new StringTokenizer(strTemp, " ", false);
        try {
            while (stringToken.hasMoreTokens()) {
                intMethod=Sip.matchMethod( (String) stringToken.
                                          nextElement());
                if (intMethod!=Sip.BadMethod) {
                    return intMethod;
                }
            }
            return Sip.BadMethod;
        }
        catch (java.util.NoSuchElementException e) {
            return Sip.BadMethod;
        }
    }

    /**
     * Generate the full message
     *
     * @return The full message text
     */
    public String message() {
        String msg;

        msg=startLine()+messageHeaders()+"\r\n";
        if (haveBody()) {
            msg+=messageBody();
        }

        return msg;
    }

    protected String messageHeaders() {
        StringBuffer headers=new StringBuffer();
        SipHeader curheader=null;

        headers.append(vialist.getViaList());

        for (ListIterator li=headerlist.listIterator(); li.hasNext(); ) {
            curheader= (SipHeader) li.next();
            if (curheader.id!=SipHeader.BadHeader) {
                headers.append(SipHeader.getHeaderString(curheader.id)+
                               ": ");
                headers.append(curheader.data+"\r\n");
            }
        }

        if (contactlist.getListLength()>0) {
            headers.append(SipHeader.getHeaderString(SipHeader.
                Contact)+": ");
            headers.append(contactlist.getUriList());
            headers.append("\r\n");
        }

        if (recordroute.getListLength()>0) {
            headers.append(SipHeader.getHeaderString(SipHeader.
                Record_Route)
                           +": ");
            headers.append(recordroute.getUriList());
            headers.append("\r\n");
        }

        return headers.toString();
    }

    protected abstract String startLine();

    protected abstract void parseStartLine(String startline);

    protected abstract void validateMessage() throws
        BadHeaderException;

    /**
     * Inserts a new header into the message.
     *
     * @param id
     * @param data
     */
    public void insertHeader(int id, String data) {
        headerlist.add(new SipHeader(id, data));
    }

    /**
     * Returns true if the message contains the specified header.
     *
     * @param id
     * @return bool
     */
    public boolean hasHeader(int id) {
        SipHeader curheader=null;

        for (ListIterator li=headerlist.listIterator(); li.hasNext(); ) {
            curheader= (SipHeader) li.next();
            if (curheader.id==id) {
                return true;
            }
        }

        return false;
    }

    /**
     * Returns the data contained in the specified header.
     *
     * @param id
     * @return String
     */
    public String getHeaderData(int id) {
        SipHeader curheader=null;

        for (ListIterator li=headerlist.listIterator(); li.hasNext(); ) {
            curheader= (SipHeader) li.next();
            if (curheader.id==id) {
                return curheader.data;
            }
        }

        return null;
    }

    /**
     * Returns a reference to the via list for this message.  Usually
     * useful for getting and setting the topmost via entry.
     *
     * @return SipViaList
     */
    public SipViaList getViaList() {
        return vialist;
    }

    /**
     * Returns a reference to the Record-Route for this message.
     *
     * @return SipUriList &
     */
    public SipUriList getRecordRoute() {
        return recordroute;
    }

    /**
     * Returns a reference to the contact list for this message.
     *
     * @return SipUriList &
     */
    public SipUriList getContactList() {
        return contactlist;
    }

    /**
         * Sets the via list to be a copy of an existing via list.  Useful for
     * building responses.
     *
     * @param copylist
     */
    public void setViaList(SipViaList copylist) {
        vialist=copylist;
    }

    /**
     * Sets the contact list for this message.
     *
     * @param newclist
     */
    public void setContactList(SipUriList newclist) {
        contactlist=newclist;
    }

    /**
     * Sets the record route for this message.
     *
     * @param newrr
     */
    public void setRecordRoute(SipUriList newrr) {
        recordroute=newrr;
    }

    /**
     * Add a url to the contact list
     */
    public void addContact(SipUri contact) {
        contactlist.add(contact);
    }

    public boolean hasSendToAddress() {
        return (sendToAddress!=null);
    }

    public SipUri getSendToAddress() {
        return sendToAddress;
    }

    public void setSendToAddress(SipUri address) {
        sendToAddress=address;
    }

    /**
     * setDefaultVars
     */
    private void setDefaultVars() {
        messageSize=0;

        vialist=new SipViaList();
        recordroute=new SipUriList();
        contactlist=new SipUriList();
        headerlist=new ArrayList();
        sendToAddress=null;
    }

    protected void parseMessage(String inputMessage) throws
        BadHeaderException {
        StringBuffer fullmessage; ;
        int wherecrlf;
        int wherecrlfcrlf;

        fullmessage=new StringBuffer(inputMessage);

        // remove \r from the message
        for (int i=0; i<fullmessage.length(); i++) {
            if (fullmessage.charAt(i)=='\r') {
                fullmessage.deleteCharAt(i);
            }
        }

        String messageString=fullmessage.toString();

        wherecrlf=messageString.indexOf("\n");
        wherecrlfcrlf=messageString.indexOf("\n\n");

        if (wherecrlfcrlf==-1) {
            wherecrlfcrlf=messageString.length();

            // Start line is the first line
        }
        parseStartLine(messageString.substring(0, wherecrlf));

        // Headers go on until \n\n
        parseHeaders(messageString.substring(wherecrlf+1,
                                             wherecrlfcrlf));

        // The body is the rest (if there is more)

        if (wherecrlfcrlf<fullmessage.length()) {
            setBody(fullmessage.substring(wherecrlfcrlf+2));

        }
        validateMessage(); // see if we've got all the appropriate headers

    }

    /**
     * parseHeaders
     * @param inbuf
     */
    protected void parseHeaders(String inbuf) throws
        BadHeaderException {
        int lastheader;
        int id;
        String curheader="";
        String curbody="";

        lastheader=0;
        HeaderParser hp=new HeaderParser(inbuf);

        while (hp.hasMoreLines()) {

            String line=hp.getLine();
            int pos=line.indexOf(':');

            if (pos<0) {
                System.err.println(
                    "HMM, in parse header with line of "+line);
                return; // nothing here
            }

            curheader=line.substring(0, pos);
            curbody=line.substring(pos+1).trim();

            // see if the line ends with a comma. If so, assume the next line(s)
            // are a continuation
            while (curbody.endsWith(",")) {
                String nextLine=hp.getLine();
                if (nextLine==null) {
                    return;
                }
                curbody+=nextLine.trim();
            }

            id=SipHeader.matchHeader(curheader);

            if (id==SipHeader.Content_Length) {
                try {
                    messageSize=Integer.parseInt(curbody);
                }
                catch (NumberFormatException nfe) {}
            }

            if (id==SipHeader.Via) {
                vialist.parseVia(curbody);
            }
            else if (id==SipHeader.Record_Route) {
                recordroute.parseList(curbody);
            }
            else if (id==SipHeader.Contact) {
                contactlist.parseList(curbody);
            }
            else {
                if (id!=SipHeader.BadHeader) {
                    insertHeader(id, curbody);
                }
            }

        }
    }
}
