/**
 * class SipRequestMessage
 *
 * Representation of a single SIP request message. SipRequestMessage is used for
 * creating and for parsing individual SIP request messages.  It also contains a
 * retransmission timer for use when sending the message.
 *
 */
package org.mitre.jsip;

import java.util.*;
import org.mitre.jsip.messages.*;

public abstract class SipRequestMessage
    extends SipMessage {
    // class variables
    private Date timestamp;
    private int lastt;
    private int sendcount;

    private static final int MAX_CALLID=Integer.MAX_VALUE-1;
    ;

    protected int meth;
    protected SipUri requesturi;

    /**
     * Static method to create a new SIP Call ID.
     *
     * @return String
     */
    public static String createCallId() {
        int uniqid;
        String theid;
        Random rand=new Random();

        uniqid=rand.nextInt(MAX_CALLID); // should be sufficient

        // fix for 3com carrier gw
        theid=String.valueOf(uniqid)+"@"+Sip.getLocalAddress();

        return theid;
    }

    /**
     * SipRequestMessage
     */
    protected SipRequestMessage() {
        super();
        setDefaultVars();
    }

    /**
     * SipRequestMessage
     */
    protected SipRequestMessage(String inputMessage) {
        super();
        setDefaultVars();

        try {
            parseMessage(inputMessage);
        }
        catch (BadHeaderException bhe) {
            System.err.println("SRM: Error parsing input: "+bhe);
        }
    }

    public static SipRequestMessage createRequestMessage(String
        inputMessage) {
        // get the method name, then create that request type
        String methodName=inputMessage.substring(0,
                                                 inputMessage.
                                                 indexOf(' '));

        SipRequestMessage message=createRequestMessage(Sip.
            matchMethod(methodName));
        if (message instanceof UnsupportedMessage) {
            return message;
        }

        message.setDefaultVars();

        try {
            message.parseMessage(inputMessage);
        }
        catch (BadHeaderException bhe) {
            System.err.println("SRM: Couldn't parse message: "+bhe);
            return null;
        }

        return message;
    }

    /**
     * Main method for getting an instance of a Request Message
     *
     * @param type The type of message to create
     * @return An instance of the message type
     */
    public static SipRequestMessage createRequestMessage(int mtype) {
        // create a message of the type requested
        switch (mtype) {
            case Sip.INVITE:
                return new InviteMessage();
            case Sip.OPTIONS:
                return new OptionsMessage();
            case Sip.REGISTER:
                return new RegisterMessage();
            case Sip.CANCEL:
                return new CancelMessage();
            case Sip.BYE:
                return new ByeMessage();
            case Sip.ACK:
                return new AckMessage();
            case Sip.MESSAGE:
                return new MessageMessage(); // bleah
            case Sip.SUBSCRIBE:
                return new SubscribeMessage();
            case Sip.NOTIFY:
                return new NotifyMessage();
        }

        // default
        return new UnsupportedMessage();
    }

    /**
     * Returns the method of this SIP request.
     *
         * @return method id as defined in <code>org.mitre.jsip.Sip</code>
     */
    public int getMethod() {
        return meth;
    }

    /**
     * Returns the type of this SIP request
     *
     * @return message type id
         /**
      * Sets the request URI for the message. This will also set the destination
      * host and port for the message.
      *
      * @param newrequri
      */
     public void setRequestUri(SipUri newrequri) {
         requesturi=newrequri;
     }

    /**
     * Returns the request URI for this message.
     *
     * @return SipUri
     */
    public SipUri getRequestUri() {
        return requesturi;
    }

    /**
     * Returns the timestamp of the last transmission.
     *
     * @return timeval *
     */
    public Date getTimestamp() {
        return timestamp;
    }

    /**
     * Recalculates the timestamp on the message.
     */
    public void setTimestamp() {
        timestamp=new Date();
    }

    /**
     * Returns the last time tick.
     *
     * @return int
     */
    public int lastTimeTick() {
        return lastt;
    }

    /**
     * Sets the current time tick.
     *
     * @param newtt
     */
    public void setTimeTick(int newtt) {
        lastt=newtt;
    }

    /**
         * Returns the current value of the retransmission counter for this
     * message.
     *
     * @return int
     */
    public int sendCount() {
        return sendcount;
    }

    /**
     * Increments the retransmission counter.
     */
    public void incrSendCount() {
        sendcount++;
    }

    /**
     * cancel this message
     */
    public void cancel() {
        meth=Sip.CANCEL;
    }

    // Need to be implemented by all subclasses
    protected abstract String startLine();

    protected abstract void validateMessage() throws
        BadHeaderException;

    /**
         * message specific logic for a new request message. The boolean return
         *  will designate whether the message has been completely handled
     *  (<code>true</code> or not (<code>false</code>)
     */
    public abstract boolean handleNewRequest(SipClient client);

    /**
     * message specific logic for a request cooresponding to a call.
     */
    public abstract void handleCallRequest(SipClient client);

    // parse the first line of the message
    protected void parseStartLine(String startline) {
        StringTokenizer st=new StringTokenizer(startline.trim());

        String method=st.nextToken();
        String requri=st.nextToken();

        meth=Sip.matchMethod(method);
        setRequestUri(new SipUri(requri));
    }

    private void setDefaultVars() {
        type=Request;
        meth=Sip.BadMethod;

        lastt=0;
        sendcount=0;
        timestamp=new Date();
    }
}
