/**
 * class SipResponseMessage
 *
 * Representation of a single SIP response message. SipResponseMessage is used for
 * creating and for parsing individual SIP response messages.
 *
 */
package org.mitre.jsip;

import java.util.*;

public class SipResponseMessage
    extends SipMessage {
    // class variables
    private SipStatus status;

    public SipResponseMessage() {
        super();
        setDefaultVars();
    }

    public SipResponseMessage(String inputMessage) {
        super();
        setDefaultVars();

        try {
            parseMessage(inputMessage);
        }
        catch (BadHeaderException bhe) {
            System.err.println("SRsM: Error parsing input: "+bhe);
        }
    }

    /**
     * Sets the status of the message.
     *
     * @param stat
     */
    public void setStatus(SipStatus stat) {
        status=stat;
    }

    /**
     * Returns the status of the message.
     *
     * @return The status of the current message
     */
    public SipStatus getStatus() {
        return status;
    }

    // set default variables
    private void setDefaultVars() {
        type=Response;
    }

    protected String startLine() {
        return getVersionString()+" "+
            String.valueOf(status.getCode())+" "+
            status.getReasonPhrase()+"\r\n";
    }

    protected void parseStartLine(String startline) {
        StringTokenizer st=new StringTokenizer(startline.trim());

        String method=st.nextToken();
        String requri=st.nextToken();
        String verstr="";

        // the verstr could actually be a response with several words
        while (st.hasMoreTokens()) {
            verstr+=st.nextToken();

        }
        try {
            status=new SipStatus(Integer.parseInt(requri));
        }
        catch (NumberFormatException nfe) {
            System.err.println("Couldn't set status code for "+requri+
                               ". Default to 500");
            status=new SipStatus(500);
        }
        status.setReasonPhrase(verstr);
    }

    protected void validateMessage() throws BadHeaderException {
        // The message should have several manditory fields. Make
        // sure they're there
        final int[] manditoryHeaders= {SipHeader.Call_ID,
                                     SipHeader.CSeq,
                                     SipHeader.From, SipHeader.To,
                                     SipHeader.Content_Length};

        for (int i=0; i<manditoryHeaders.length; i++) {
            // check for their existance. If not there, throw the exception
            if (!hasHeader(manditoryHeaders[i])) {
                throw new BadHeaderException(
                    "Message is missing header: "+
                    SipHeader.getHeaderString(manditoryHeaders[i]));
            }
        }

        // make sure there's a "Via"
        if (vialist.size()==0) {
            throw new BadHeaderException(
                "Message is missing header: "+
                SipHeader.getHeaderString(SipHeader.Via));
        }

        // a couple of unique checks

        // All 405 responses should have an Allow field
        if (status.getCode()==405) {
            if (!hasHeader(SipHeader.Allow)) {
                throw new BadHeaderException(
                    "Message is missing header: "+
                    SipHeader.getHeaderString(SipHeader.Allow));
            }
        }
    }

}
