/*
 * This file is part of the Java SIP library, an open source SIP library. Please see
 * http://collaboration.mitre.org/sip for more information
 *
 * Copyright (c) 2001 The MITRE Corporation
 *
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 */
/**
 * Test class for the Sip library
 *
 * @author S.R.Jones
 * @version 1
 */
package org.mitre.jsip;

import java.io.*;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import org.mitre.jsip.event.*;

public class SipTester extends JFrame implements Runnable {

    private SipClient client;
    private SipUser  localUser;

    private SipRegister regServer;

    private static final boolean DEBUG = true;

    // various testing gui variables
    private JRadioButton regButton;
    private JRadioButton invButton;

    private JTextArea scrollback;
    private JTextField entry;

    // used in test messages
    private final String remoteUri = "sip:srjones@blabla.mitre.org";
    private final String regServerUri = "sip:sturgis.mitre.org";
    private final int [] payloadTypes = {0, 3};
    private final String mcastAddr = "225.5.5.5";
    private final int mcastPort = 23322;

    // MESSAGE testing
    private static String imUri;
    private SipCall call;
    private SipCallMember call_leg;

    // Presence testing
    //    private PresenceAgent presenceAgent;

    public SipTester (String contact) {
	// try initializing the same way kphone does
	SipUri uri = new SipUri( contact );
	client = SipClient.getSipClient();
	localUser = null;

	client.setDefaultUserMode( true );
	client.setExplicitProxyMode( false );

	String laddr = Sip.getLocalAddress();
	client.setCallForward( false );
	client.setCallForwardUri( null );
	client.setCallForwardMessage( null );
	client.setMaxForwards( 0 );
	client.setBusyMessage( null );
	client.setHideViaMode( SipClient.HideHop );

	localUser = new SipUser( client, uri );
	client.setDefaultUser( localUser );
	SipUri remoteuri = new SipUri( imUri );
	call = new SipCall( localUser, null );
	call_leg = new SipCallMember( call, remoteuri );

	// init the presence agents (for IM testing)
	//	initPresenceAgents();

	// build the gui
	//	initGui();
	initSelectGui();

	// add the listeners for incoming messages
	CallListener cl =  new CallListener() {
		public void responseReceived(CallEvent ce) {
		    // do more with this later
		    System.out.println("Received response message event");
		}

		public void requestReceived(CallEvent ce) {
		    // same with this
		    System.out.println("ST: Received request message event");
		    SipRequestMessage requestMessage = (SipRequestMessage) ce.getMessage();
		    handleRequestReceived( requestMessage );
		}
	    };
	client.addCallListener(cl);

	if ( DEBUG ) System.out.println( "SipClient: initialization complete" );
    }

    protected void initGui() {

	Container cp = this.getContentPane();
	cp.setLayout( new BorderLayout() );

	scrollback = new JTextArea(10, 40); // 10 rows and 40 columns
	cp.add( new JScrollPane( scrollback ), BorderLayout.CENTER );

	JPanel entryPanel = new JPanel();
	entry = new JTextField( 40 );
	JButton sendButton = new JButton( "Send" );

	sendButton.addActionListener( new ActionListener() {
		public void actionPerformed( ActionEvent ae ) {
		    // get the text in the entry field, then clear out entry field
		    String message = entry.getText();
		    entry.setText("");

		    // put the message in Text Area, then send out
		    scrollback.append( localUser.getUri().user() + ": " + message + "\n" );
		    sendMessage( message );
		}
	    });

	entryPanel.add( entry );
	entryPanel.add( sendButton );

	cp.add( entryPanel, BorderLayout.SOUTH );
    }

    protected void initSelectGui() {
	Container cp = this.getContentPane();
	cp.setLayout( new BorderLayout() );

	JPanel centerPanel = new JPanel( new BorderLayout() );
	ButtonGroup bg = new ButtonGroup();
	regButton = new JRadioButton();
	regButton.setText( "Send REGISTER" );
	regButton.setSelected( true );
        invButton = new JRadioButton();
	invButton.setText( "Send INVITE" );
	bg.add(regButton);
	bg.add(invButton);

	centerPanel.add(regButton, BorderLayout.NORTH);
	centerPanel.add(invButton, BorderLayout.CENTER);

	JPanel buttonPanel = new JPanel();
	JButton okButton = new JButton();
	JButton cancelButton = new JButton();

	okButton.setText( "Execute" );
	okButton.addActionListener( new ActionListener() {
		public void actionPerformed(ActionEvent evt) {
		    executeCommand();
		}
	    });

	cancelButton.setText( "Exit" );
	cancelButton.addActionListener( new ActionListener() {
		public void actionPerformed( ActionEvent evt ) {
		    exitCommand();
		}
	    });

	buttonPanel.add(okButton);
	buttonPanel.add(cancelButton);

	cp.add( centerPanel, BorderLayout.CENTER );
	cp.add( buttonPanel, BorderLayout.SOUTH );

	setSize( 240, 150 );
    }

    protected void handleRequestReceived( SipRequestMessage requestMessage ) {
	SipUri sender = new SipUri( requestMessage.getHeaderData( SipHeader.From ) );

	if ( sender == null ) {
	    System.err.println( "Don't know where this came from... tossing" );
	    return;
	}

	// real system would handle different messages correctly
	// however, since this is a test, just doing IM for now
	if ( requestMessage.getMethod() == Sip.MESSAGE ) {
	    // put it up on the text area
	    String message = requestMessage.messageBody();
	    System.err.println("Message received is: [" + message + "]");

	    String newText = sender.user() + ": " + message + "\n";
	    scrollback.append( newText );
	}

    }

//      public void startPresenceListener() {
//  	// create a listener for PresenceEvents
//  	PresenceListener pl = new PresenceListener() {
//  		public void subscribeReceived( PresenceEvent pe ) {
//  		    System.out.println("Received subscribe message");
//  		    handleSubscribeMessage( pe );
//  		}

//  		public void notifyReceived( PresenceEvent pe ) {
//  		    System.out.println("Received Notify message");
//  		}
//  	    };
//  	SipClient.getSipClient().addPresenceListener( pl );

//      }

//      protected void handleSubscribeMessage( PresenceEvent pe ) {
//  	// create a popup dialog to confirm request
//  	SipMessage incomingMessage = pe.getMessage();

//  	String query = "You received a Subscribe request from \n" +
//  	    incomingMessage.getHeaderData( SipHeader.From ) +
//  	    "\nWould you like to allow this person to see your presence?";

//  	int rc = JOptionPane.showConfirmDialog( this, query,
//  						"Subscribe Request",
//  						JOptionPane.YES_NO_OPTION );

//  	if ( rc == JOptionPane.OK_OPTION ) {
//  	    // send an ok first
//  	    client.sendQuickResponse( incomingMessage, new SipStatus( 202 ) );

//  	    // we should do something with the Contact header....
//  	    // probably not the right way.. will check this more
//  	    // add this user to the client's user list first
//  	    Presentity newUser = new Presentity( incomingMessage.getHeaderData( SipHeader.From ) );
//  	    //	    client.addUser( new SipUser( client, newUser ) );
//  	    presenceAgent.addUser( newUser );
//  	} else {
//  	    // rejected!
//  	    client.sendQuickResponse( incomingMessage, new SipStatus( 403 ) ); //????
//  	}
//      }

    public void run() {
	// run a loop in here to handle messages
	for (;;) {
	    while (!client.haveMessages() ) {
		try {
		    Thread.currentThread().sleep(3000);
		    //		    System.out.print(".");
		} catch (InterruptedException ie) {break;}
	    }

	    System.out.println("Message in queue, forwarding...");
	    client.incomingMessage();
	}
    }

    protected void sendRegister(String username, String password) {

	// create a listener for REGISTER responses
	CallListener regCallListener = new CallListener() {
		public void responseReceived(CallEvent ce) {
		    System.out.println( "REGCALLLISTENER activated" );

		    // see if this is a response message
		    SipResponseMessage respMessage = (SipResponseMessage)ce.getMessage();
		    if ( respMessage.getStatus().getCode() == 401 ) {
			System.out.println("REGCALLLISTENER - updating status");
			try {
			    regServer.setAuthentication(respMessage);
			} catch ( RegisterException re ) {
			    System.err.println( re.toString() );
			    return;
			}
			try {
			    regServer.requestRegister( false );   // try to register again
			} catch ( RegisterException re ) {
			    System.err.println("Couldn't register: " + re );
			    return;
			}
		    }
		}

		public void requestReceived(CallEvent ce) {
		    // ignore this
		}
	    };
	SipClient.getSipClient().addCallListener( regCallListener );

	// create a REGISTER message to a server
	regServer = localUser.addServer( regServerUri );

	try {
	    regServer.requestRegister( password );
	} catch ( RegisterException re ) {
	    System.err.println( re );
	}
    }

    protected void sendMessage( String message ) {
	call_leg.requestMessage( message );
    }

    protected void sendInvite() {
	// create an INVITE message and send it out
	SdpMessage sdpMessage = new SdpMessage();
	sdpMessage.setName("jSIP");
	sdpMessage.setUsername("steve");
	sdpMessage.addMedia( SdpMessage.AUDIO, "0", "RTP/AVP", mcastPort);
	sdpMessage.setIpAddress(mcastAddr);
	sdpMessage.setPort(mcastPort);

	SipUri remoteuri = new SipUri(remoteUri);
	SipCall call = new SipCall( localUser, null );    // this is for creating a brand new call
	SipCallMember member = new SipCallMember( call, remoteuri );

	member.requestInvite( sdpMessage.message(), new MimeContentType( "application/sdp" ) );
    }

    protected void executeCommand() {
	if ( regButton.isSelected() ) {
	    sendRegister("user1", "none"); // junk
	} else if ( invButton.isSelected() ) {
	    sendInvite();
	} else {
	    System.out.println("Don't know which command to execute");
	}
    }

    public void exitCommand() {
	// try to unregister from the registry server

	try {
	    regServer.requestRegister( true );
	} catch ( RegisterException re ) {}

	System.exit( 0 );
    }

//      // this initializes the presence and watcher agents
//      protected void initPresenceAgents() {
//  	// init the presence agent. This tracks subscriptions you've ok'ed
//  	System.setProperty( "jsip.presence.subscriptions",
//  			    "/tmp/" + localUser.getUri().uri() );
//  	presenceAgent = new PresenceAgent( new Presentity( localUser.getUri().uri() ) );
//  	presenceAgent.setPresence( PresenceAgent.PRESENCE_AVAILABLE );

//  	// init the watcher agent. This tracks your buddies
//  	// TBD (to be done)
//      }

    public static void main(String[] args) {
	// testing IM stuff
	String reginfo = args[0];   // this consists of username:password
	imUri = args[1];

	int colon = reginfo.indexOf( ':' );
	if ( colon < 0 ) {
	    System.err.println( "First argument should be in form username:password" );
	    System.exit( 1 );
	}
	String contact = reginfo.substring( 0, colon );
	String password = reginfo.substring( colon + 1 );

	final SipTester tester = new SipTester( contact );
	tester.setTitle( "Chat with " + imUri );

	// start the message queue thread
	Thread messageThread = new Thread(tester);
	messageThread.start();

	tester.pack();
	tester.setVisible( true );

	// exit the app when the window closes
	tester.addWindowListener( new WindowAdapter() {
		public void windowClosed( WindowEvent e ) {
		    tester.exitCommand();
	    //		    messageThread.stop();
		}
	    });

	// try to register now
	// Comment this out if you don't have a Register server
	tester.sendRegister( contact, password );

	//	tester.startPresenceListener();
    }
}

