/*
 * This file was derived from libdissipate, an open source SIP library. The original file
 * was modified on 1/23/2001.  Please see
 * http://www.div8.net/dissipate for more information.
 *
 * Copyright (c) 2000 Billy Biggs <bbiggs@div8.net>
 *
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to the
     * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 */
/**
 * class SipTransaction
 *
 * This code has been generated using C2J++
 * C2J++ is based on Chris Laffra's C2J (laffra@watson.ibm.com)
 * Read general disclaimer distributed with C2J++ before using this code
 * For information about C2J++, send mail to Ilya_Tilevich@ibi.com
 *
 *change Log
 *------------------------------------------------
 *Author         Version    Date       Description
 *A.N Sudhakar   00.00.02   05-01-2001 Made Response to throw events for Response messages of all Requests
 *A.N Sudhakar   00.00.01   26-12-2001 Took care of Transaction which was taking 110 seconds
 */
package org.mitre.jsip;

import java.util.ArrayList;
import java.util.Date;
import java.util.Timer;
import java.util.TimerTask;

import org.mitre.jsip.event.*;

public class SipTransaction
    extends Object {
    // class variables
    static final int RemoteRequest=0;
    static final int LocalRequest=1;
    static final int None=2;

    private static final String DEFAULT_EXPIRATION="3600";

    private SipCall parent;
    private SipRequestMessage requestmessage;
    private ArrayList responseList;
    private SipStatus laststatus;
    private int seqnum;
    private SipCallMember remote;
    private SipUri contacturi;
    private int direction;
    private boolean cancelled;
    private boolean gotAck;

    // use a single Timer to schedule retransmission checks
    private static Timer retransmissionTimer;

    // instances of this class that execute requests will track if retransmissions are necessary
    private TimerTask retransmissionTimerTask;

    // how often we should run the check (in ms)
    private static final int RETRANS_TIMER_CHECK=3005;

    /**
     * SipTransaction
     * @param seqn
     * @param farend
     * @param call
     */
    public SipTransaction(int seqn, SipCallMember farEnd,
                          SipCall call) {
        responseList=new ArrayList();
        parent=call;
        requestmessage=null;
        direction=None;
        remote=farEnd;
        seqnum=seqn;
        cancelled=false;
        gotAck=false;
    }

    /**
     * getCallMember
     * @return SipCallMember *
     */
    public SipCallMember getCallMember() {
        return remote;
    }

    /**
     * getCSeq
     * @return String
     */
    public String getCSeq() {
        String cseq=String.valueOf(seqnum)
                    +" "+Sip.getMethodString(requestmessage.getMethod());

        return cseq;
    }

    public boolean isAcked() {
        return gotAck;
    }

    public void Acked() {
        gotAck=true;
    }

    /**
     * getDirection
     * @return Direction
     */
    public int getDirection() {
        return direction;
    }

    /**
     * getSeqNum
     * @return int
     */
    public int getSeqNum() {
        return seqnum;
    }

    /**
     * getStatus
     * @return SipStatus
     */
    public SipStatus getStatus() {
        return laststatus;
    }

    /**
     * wasCancelled
     * @return bool
     */
    public boolean wasCancelled() {
        return cancelled;
    }

    /**
     * getRequest
     * @return SipMessage *
     */
    public SipRequestMessage getRequest() {
        return requestmessage;
    }

    public void sendResponse(SipStatus status) {
        sendResponse(status, null, null);
    }

    /**
     * sendResponse
     */
    public void sendResponse(SipStatus status, String body,
                             MimeContentType bodytype) {
        parent.sendResponse(constructResponse(status, body, bodytype));
    }

    private SipResponseMessage constructResponse(SipStatus status,
                                                 String body,
                                                 MimeContentType
                                                 bodytype) {
        SipResponseMessage response=new SipResponseMessage();
        setStatus(status);
        responseList.add(response);

        response.setStatus(status);

        response.setViaList(requestmessage.getViaList());

        response.insertHeader(SipHeader.From,
                              remote.getUri().nameAddr());
        response.insertHeader(SipHeader.CSeq, getCSeq());

        if (bodytype!=null) {
            response.insertHeader(SipHeader.Content_Type,
                                  bodytype.type());
        }

        response.setBody(body);
        return response;
    }

    /**
     * cancelRequest
     */
    public void cancelRequest() {
        cancelRequest(null, null);
    }

    public void cancelRequest(String body, MimeContentType bodytype) {
        requestmessage.cancel();

        // maybe we should have a sendRequest that takes the requestmessage itself?
        // -- sendRequest here is going to rebuild requestmessage ;-p
        sendRequest(Sip.CANCEL, body, bodytype);
    }

    /**
     * getFinalMessageBody
     * @return String
     */
    public String getFinalMessageBody() {
        int index=responseList.size();
        if (index>0) {
            SipMessage sm= (SipMessage) responseList.get(index-1);
            return sm.messageBody();
        }

        return null;
    }

    /**
     * getFinalContentType
     * @return MimeContentType
     */
    public MimeContentType getFinalContentType() {
        int index=responseList.size();
        if (index>0) {
            SipMessage sm= (SipMessage) responseList.get(index-1);
            return new MimeContentType(sm.getHeaderData(SipHeader.
                Content_Type));
        }

        return new MimeContentType();
    }

    /**
     * getFinalContactList
     * @return SipUriList
     */
    public SipUriList getFinalContactList() {
        int index=responseList.size();
        if (index>0) {
            SipMessage sm= (SipMessage) responseList.get(index-1);
            return sm.getContactList();
        }

        return null;
    }

    /**
     * getFinalWWWAuthString
     * @return String
     */
    public String getFinalWWWAuthString() {
        int index=responseList.size();
        if (index>0) {
            SipMessage sm= (SipMessage) responseList.get(index-1);
            return sm.getHeaderData(SipHeader.WWW_Authenticate);
        }

        return null;
    }

    /**
     * getFinalProxyAuthString
     * @return String
     */
    public String getFinalProxyAuthString() {
        int index=responseList.size();
        if (index>0) {
            SipMessage sm= (SipMessage) responseList.get(index-1);
            return sm.getHeaderData(SipHeader.Proxy_Authenticate);
        }

        return null;
    }

    /**
     * getRequestMessageBody
     * @return String
     */
    public String getRequestMessageBody() {
        return requestmessage.messageBody();
    }

    /**
     * getRequestMessageContentType
     * @return MimeContentType
     */
    public MimeContentType getRequestMessageContentType() {
        if (requestmessage.hasHeader(SipHeader.Content_Type)) {
            return new MimeContentType(requestmessage.getHeaderData(
                SipHeader.Content_Type));
        }

        return null;
    }

    /**
     * statusUpdated
     */
    //private void statusUpdated() {
    //SipCall.statusUpdated();  //*** LOOK AT ***
     //}

     /**
      * setStatus
      * @param stat
      */
     private void setStatus(SipStatus stat) {
         laststatus=stat;
         //statusUpdated();
         remote.remoteStatusUpdated(); // ????
     }

    /**
     * incomingRequest
     * @param message
     */
    void incomingRequest(SipRequestMessage message) {
        SdpMessage optsdp=new SdpMessage();

        if (SipClient.DEBUG) {
            System.out.println("SipTransaction: Incoming Request");

        }
        if ( (message.getMethod()==Sip.ACK)) {
            if (SipClient.DEBUG) {
                System.out.println(
                    "SipTransaction: Received an ACK...");

                // Make sure listeners know in case there was a new session description
            }
            remote.remoteStatusUpdated(); // ????

        }
        else if ( (message.getMethod()==Sip.CANCEL)) {
            if (SipClient.DEBUG) {
                System.out.println(
                    "SipTransaction: CANCEL Received, sending an OK");

            }
            requestmessage.cancel();
            sendResponse(new SipStatus(200));

            // Mark request as cancelled and inform listeners
            cancelled=true;
            remote.remoteStatusUpdated();

        }
        else {
            direction=RemoteRequest;
            requestmessage=message;
            switch (message.getMethod()) {
                case Sip.INVITE:
                    parent.setCallType(SipCall.StandardCall);
                    sendResponse(new SipStatus(100));
                    break; // changed from a return right now
                case Sip.BYE:
                    if (parent.getCallType()==SipCall.UnknownCall) {
                        parent.setCallType(SipCall.BrokenCall);
                    }

                    //sendResponse( new SipStatus( 200 ) );
                    break;
                case Sip.OPTIONS:
                    if (parent.getCallType()==SipCall.UnknownCall) {
                        parent.setCallType(SipCall.OptionsCall);
                    }

                    optsdp.setIpAddress("0.0.0.0");
                    optsdp.setName("");
                    optsdp.setPort(0);

                    sendResponse(new SipStatus(200), optsdp.message(),
                                 new
                                 MimeContentType("application/sdp"));
                    break;
                case Sip.MESSAGE:
                    parent.setCallType(SipCall.MessageCall);
                    if (message.hasHeader(SipHeader.Subject)) {
                        parent.setSubject(message.getHeaderData(
                            SipHeader.Subject));
                    }
                    sendResponse(new SipStatus(200));
                    break;
                default:
                    break;
            }
        }
        // send the request event
        // LOOK AT???
        //  	CallListener callListener = SipClient.getSipClient().getCallListener();
        CallEventMulticaster cemc=CallEventMulticaster.getMulticaster();
        CallEvent ce=new CallEvent(this,
                                   message,
                                   CallEvent.CALLREQUEST_ID);
        //	callListener.requestReceived( ce );
        cemc.requestReceived(ce);
    }

    /**
     * incomingResponse
     * @param message
     */
    void incomingResponse(SipResponseMessage message) {
        if (SipClient.DEBUG) {
            System.out.println("SipTransaction: Incoming Response\n");

            // stop the retransmit timer
        }
        stopTimer();

        responseList.add(message);

        // Grab tags
        SipUri newto=new SipUri(message.getHeaderData(SipHeader.To)); // *** LOOK AT ****
        if (newto.hasTag()) {
            remote.setUri(newto);
        }

        // We have a response to our request. If necessary, we'll generate a CallEvent to notify the user
        // (when they answer, this in turn will change the local status)
        if (SipClient.DEBUG) {
            System.out.println("Sending CallResponseEvent");
//  	if( ( requestmessage.getMethod() == Sip.INVITE ) ||
//  	    ( requestmessage.getMethod() == Sip.REGISTER ) )
//  	    {
            // the far end agreed to meet. Signal the user
            //	CallListener callListener = SipClient.getSipClient().getCallListener();
        }
        CallEventMulticaster cemc=CallEventMulticaster.getMulticaster();
        CallEvent ce=new CallEvent(this,
                                   message,
                                   CallEvent.CALLRESPONSE_ID);
        cemc.responseReceived(ce);

//    	}

        if ( (requestmessage.getMethod()==Sip.INVITE)&&
            (message.getStatus().getCode()>=200)) {
            sendRequest(Sip.ACK);
        }

        setStatus(message.getStatus());
    }

    /**
     * sendRegister
     */
    void sendRegister(SipUri registerserver, boolean clear,
                      String authentication,
                      String proxyauthentication, String body,
                      MimeContentType bodytype) {
        if (requestmessage==null) {
            direction=LocalRequest;

        }
        requestmessage=SipRequestMessage.createRequestMessage(Sip.
            REGISTER);

        // set request URI
        requestmessage.setRequestUri(registerserver);
        requestmessage.insertHeader(SipHeader.CSeq, getCSeq());
        requestmessage.insertHeader(SipHeader.To,
                                    remote.getUri().nameAddr());

        if (authentication!=null) {
            requestmessage.insertHeader(SipHeader.Authorization,
                                        authentication);

        }
        if (proxyauthentication!=null) {
            requestmessage.insertHeader(SipHeader.Proxy_Authorization,
                                        proxyauthentication);

        }
        if (bodytype!=null) {
            requestmessage.insertHeader(SipHeader.Content_Type,
                                        bodytype.type());

        }
        requestmessage.setBody(body);

        if (clear) {
            requestmessage.insertHeader(SipHeader.Expires, "0");
        }
        else {
            requestmessage.insertHeader(SipHeader.Expires,
                                        DEFAULT_EXPIRATION);
        }

        parent.sendRequest(requestmessage, true);
        startTimer(); // this handles retransmissions
    }

    /**
     * sendRequest
     */
    void sendRequest(int method) {
        sendRequest(method, null, null, null);
    }

    void sendRequest(int method, String body,
                     MimeContentType bodytype) {
        sendRequest(method, body, bodytype, null);
    }

    void sendRequest(int method, String body,
                     MimeContentType bodytype, SipUri transferto) {
        if (method==Sip.INVITE) {
            parent.setCallType(SipCall.StandardCall);

        }
        if (requestmessage==null) {
            direction=LocalRequest;

            // construct message
        }
        requestmessage=SipRequestMessage.createRequestMessage(method);

        // set request uri
        requestmessage.setRequestUri(remote.getContactUri());
        requestmessage.insertHeader(SipHeader.CSeq, getCSeq());

        //if ( method == Sip.MESSAGE )
        //	    remote.getUri().setProtocolName( "im" );

        requestmessage.insertHeader(SipHeader.To,
                                    remote.getUri().nameAddr());

        // figure out the contact info **** LOOK AT **** why is this here????
//  	SipUri contact = new SipUri();
//  	contact.setUsername(remote.getUri().getUsername());
//  	contact.setHostname(Sip.getLocalAddress());
//  	contact.setPortNumber(Sip.getLocalPort());

//  	requestmessage.addContact( contact );

        if (transferto!=null) {
            requestmessage.insertHeader(SipHeader.Also,
                                        transferto.nameAddr());

        }
        if (bodytype!=null) {
            requestmessage.insertHeader(SipHeader.Content_Type,
                                        bodytype.type());

            // set event to presence if this is a SUBSCRIBE or NOTIFY
            // what are other events?
        }
        if (requestmessage.getMethod()==Sip.SUBSCRIBE||
            requestmessage.getMethod()==Sip.NOTIFY) {
            requestmessage.insertHeader(SipHeader.Event, "presence");
        }

        requestmessage.setBody(body);

        parent.sendRequest(requestmessage);
        startTimer(); // this handles retransmissions
    }

    private void startTimer() {
        // initialize
        laststatus=new SipStatus(0);

        // make sure the timer has been created
        if (retransmissionTimer==null) {
            retransmissionTimer=new Timer();

            // create the task to check for when retransmissions are required
        }
        retransmissionTimerTask=new TimerTask() {
            public void run() {
                auditPending();
            }
        };

        // schedule this for the specified time
        retransmissionTimer.schedule(retransmissionTimerTask, 0,
                                     RETRANS_TIMER_CHECK);
    }

    private void stopTimer() {
        if (retransmissionTimerTask!=null) {
            retransmissionTimerTask.cancel();
        }
    }

    /**
     * auditPending
     */
    void auditPending() {
        Date tv;
        Date reftv;
        long diff;

        if ( (laststatus.getCode()<200)&& (direction==LocalRequest)) {

            // Check time
            tv=new Date();
            reftv=requestmessage.getTimestamp();

            // Check difference
            diff=tv.getTime()-reftv.getTime();

            if (requestmessage.getMethod()!=Sip.INVITE) {
                if (diff>requestmessage.lastTimeTick()) {

                    // Timeout
                    if (requestmessage.sendCount()==11) {
                        setStatus(new SipStatus(408));
                        SipResponseMessage timeout=constructResponse(
                            laststatus, null, null);

                        CallEventMulticaster cemc=
                            CallEventMulticaster.getMulticaster();
                        CallEvent ce=new CallEvent(this, timeout,
                            CallEvent.CALLREQUEST_ID);
                        cemc.responseReceived(ce);
                        cemc.dispatchEvents();

                        return;
                    }
                    System.err.println("SipTransaction: Retransmit "+
                                       requestmessage.sendCount()+
                                       " ("
                                       +requestmessage.lastTimeTick()+
                                       ")");

                    parent.sendRaw(requestmessage);
                    requestmessage.setTimestamp();
                    requestmessage.incrSendCount();

                    if (requestmessage.lastTimeTick()<4000) {
                        if (laststatus.getCode()<100) {
                            requestmessage.setTimeTick(2*
                                requestmessage.lastTimeTick());
                        }
                        else {
                            requestmessage.setTimeTick(4000);
                        }
                    }
                }
            }
            else {
                // INVITEs should be retransmitted differently
                if (laststatus.getCode()<100) {
                    if (requestmessage.sendCount()==7) {
                        if (SipClient.DEBUG) {
                            System.out.println(
                                "SipTransaction: Request Timed Out at "
                                +requestmessage.sendCount());
                        }
                        setStatus(new SipStatus(408));
                        SipResponseMessage timeout=constructResponse(
                            laststatus, null, null);
                        CallEventMulticaster cemc=
                            CallEventMulticaster.getMulticaster();
                        CallEvent ce=new CallEvent(this, timeout,
                            CallEvent.CALLREQUEST_ID);
                        cemc.responseReceived(ce);
                        cemc.dispatchEvents();

                        return;
                    }

                    if (diff>requestmessage.lastTimeTick()) {
                        if (SipClient.DEBUG) {
                            System.out.println(
                                "SipTransaction: Retransmit "
                                +requestmessage.sendCount()+" ("
                                +requestmessage.lastTimeTick()+")");

                        }
                        parent.sendRaw(requestmessage);
                        requestmessage.setTimestamp();
                        requestmessage.incrSendCount();

                        requestmessage.setTimeTick(2*
                            requestmessage.lastTimeTick());
                    }
                }
            }
        }
    }
}
